// g++ -Wall -O3 -msse2 -o mandelbulb_gpu -lGL $($SDL13_DIR/bin/sdl-config --cflags --libs) mandelbulb_gpu.cpp
#include <iostream>
#include <fstream>
#include <cmath>
#include <SDL.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>

const float MIN_X = -1.1f;
const float MAX_X = 1.1f;
const float MIN_Y = -1.1f;
const float MAX_Y = 1.1f;
const float MIN_Z = -1.1f;
const float MAX_Z = 1.1f;
const unsigned STEPS_X = 1000;
const unsigned STEPS_Y = 1000;
const unsigned STEPS_Z = 1000;
const float DX = (MAX_X-MIN_X)/static_cast<float>(STEPS_X);
const float DY = (MAX_Y-MIN_Y)/static_cast<float>(STEPS_Y);
const float DZ = (MAX_Z-MIN_Z)/static_cast<float>(STEPS_Z);
const unsigned MAX_ITERS = 6;
const float BAILOUT = 4.0;

void shader_log(const char* msg,
                GLuint shader)
{
  GLint log_length = 0;
  glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &log_length);
  if ( log_length > 0 )
  {
    GLchar* log = new GLchar[log_length];
    std::fill(log, log+log_length, '\0');
    GLsizei real_length = 0;
    glGetShaderInfoLog(shader,
                       log_length,
                       &real_length,
                       log);
    if ( real_length > 0 )
    {
      std::cout << msg << std::endl << log << std::endl;
    }
    delete[] log;
  }
}
void read_shader(const std::string& file,
                 std::string& src)
{
  std::ifstream ifs(file.c_str());
  while ( not ifs.eof() )
  {
    std::string line;
    std::getline(ifs, line);
    src.append(line);
    src.append("\n");
  }
  ifs.close();
}

int main(int argc, char* argv[])
{
  SDL_Init(SDL_INIT_VIDEO);
  // initialise OpenGL context
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, 32);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
  SDL_WindowID fenetre =
    SDL_CreateWindow("Mandelbuld GPU (shader)",
                     SDL_WINDOWPOS_CENTERED,
                     SDL_WINDOWPOS_CENTERED,
                     STEPS_X, STEPS_Y,
                     SDL_WINDOW_OPENGL);
  SDL_GLContext context = SDL_GL_CreateContext(fenetre);
  SDL_GL_SetSwapInterval(1);
	// load and compile vertex shader
  GLuint vertex_shader = glCreateShader(GL_VERTEX_SHADER);
  {
    std::string shader_str;
    read_shader("mandelbulb_vertex_shader.glsl",
                shader_str);
    const GLchar* shader_src = shader_str.c_str();
    glShaderSource(vertex_shader,
                    1,
                    &shader_src,
                    0);
  }
  glCompileShader(vertex_shader);
  shader_log("Vertex shader log:", vertex_shader);
	// load and compile fragment shader
  GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
  {
    std::string shader_str;
    read_shader("mandelbulb_fragment_shader.glsl",
                shader_str);
    const GLchar* shader_src = shader_str.c_str();
    glShaderSource(fragment_shader,
                    1,
                    &shader_src,
                    0);
  }
  glCompileShader(fragment_shader);
  shader_log("Fragment shader log:", fragment_shader);
	// create program
  GLuint shader_program = glCreateProgram();
  glAttachShader(shader_program, vertex_shader);
  glAttachShader(shader_program, fragment_shader);
  glLinkProgram(shader_program);
  glUseProgram(shader_program);
	// set limits
  glUniform1f(glGetUniformLocation(shader_program, "BAILOUT"), BAILOUT);
  glUniform1i(glGetUniformLocation(shader_program, "MAX_ITERS"), MAX_ITERS);
	// setup view
  glViewport(0, 0, STEPS_X, STEPS_Y);
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glOrtho(MIN_X, MAX_X, MIN_Y, MAX_Y, 1.0, -1.0);
	// compute Mandelbulb's density file
  unsigned char* plane = new unsigned char[STEPS_X*STEPS_Y];
  FILE* f;
  f = fopen((std::string(argv[0])+".df3").c_str(), "wb");
  fprintf(f, "%c%c%c%c%c%c",
          STEPS_X >> 8, STEPS_X & 0xFF,
          STEPS_Y >> 8, STEPS_Y & 0xFF,
          STEPS_Z >> 8, STEPS_Z & 0xFF);
  const unsigned long long ticks_start = SDL_GetTicks();
  const GLuint z_level_loc = glGetUniformLocation(shader_program, "Z_LEVEL");
  for(unsigned iz = 0; iz < STEPS_Z; ++iz)
  {
    std::cout << "Plane " << iz << "/" << STEPS_Z << "...\r";
    std::cout.flush();
    const float cz = MIN_Z + (static_cast<float>(iz)+0.5f)*DZ;
    glUniform1f(z_level_loc, cz);
    glClear(GL_COLOR_BUFFER_BIT bitor GL_DEPTH_BUFFER_BIT);
    glBegin(GL_QUADS);
      glVertex2f(MIN_X, MIN_Y);
      glVertex2f(MAX_X, MIN_Y);
      glVertex2f(MAX_X, MAX_Y);
      glVertex2f(MIN_X, MAX_Y);
    glEnd();
    glFlush();
    glReadPixels(0, 0, STEPS_X, STEPS_Y, GL_LUMINANCE, GL_UNSIGNED_BYTE, plane);
    fwrite(plane, 1, STEPS_X*STEPS_Y, f);
  }
  std::cout << std::endl;
  const unsigned long long ticks_end = SDL_GetTicks();
  std::cout << "Elapsed: " << (ticks_end-ticks_start) << "ms\n";
  fclose(f);
  delete[] plane;
  SDL_GL_DeleteContext(context);
  SDL_DestroyWindow(fenetre);
  SDL_Quit();
	std::cout << std::string(argv[0])+".df3" << " saved." << std::endl;
  return 0;
}
